# cfhtfirst.cl -- CFHT/KIR specific version of first
#
# 1. hedit image headers, adding: 
# 	1. BPM,BPMASK = bad pixel mask name
# 	2. ut,st     = Universal & Sidereal time (copied from UTIME & SIDTIME)
#	3. coadds   = number of internal coadds (from RDTMODE)
# 	4. exptime  = (inttime: exposure time per coadd) * (# of coadds)
# 	5. obs-date = old date-obs but in dd/mm/yy format
# 	6. date-obs = new millenium format (old date-obs saved as dateorig)
# 2. subtract dark
#
# 940902: fixed problem with ut>24
# 950122: adjusted so that header params are added first, to raw image, and
#   dark is then subtracted, so that both sets of images have header params.
# Also added `DARKSUB' header card to darksubbed images, indicating what
#   dark image was subtracted from them.
# Also fixed problem with BPM not getting added into image header.
# 950203: added adding of "coadds" header card
# 950214: check for "ncoadds" header card & ok date-obs (laird's fits256.e data)
# 950215: use "obs-date" to find day, month, year since it has standard format;
#   also streamlined input parameters.
# 950216: made adding of airmass optional
# 970321: yyear calculated correctly for observations from 1981-2080 only
# 980414: added compatibility to incoming date-obs in new-millenium format and
#   output of date-obs in same format (valid for data from 1990 to 2089);
#   original date-obs is output as dateorig
# 981228: adapted to CFHT + KIR: removed airmass & coords options;
#   now adds UT directly from UTIME and ST from SIDTIME; 
# 981229: added optional step to convert from default 3-D format to 2-D
# 990427: DARKSUB keyword added to darksubbed images, not original ones
#
# **should add ability to do nonlinearity correction if desired
# **assumes that INTTIME is exposure time per coadd, which should be correct

procedure cfhtfirst(imagelist,darksub,badmask,project)

string imagelist  {prompt="List of images to process (do not use an `@')"}
bool   darksub    {prompt="Subtract off dark frame?"}
string badmask    {prompt="Name of bad pixel mask to add to image header"}
bool   project=yes {prompt="Convert default 3-D image format to 2-D?"}

string dark       {prompt="Name of the dark frame to subtract", mode="q"}
string darkstring {prompt="String to append to dark-subtracted images",mode="q"}
string observatory="cfht" {prompt="Observatory where observations were made"}
bool   ddmmyy=yes  {prompt="Is 'date-obs' in day/month/year format?"}
bool   echog=yes   {prompt="Beep when task is done?"}
struct *list

begin

string ddark, ddarkstring, bbadmask, check, base, iimagelist, ext
string darklist, flat, flatbase, meanfile, currentflat, tempflat, name
string editlist, current, obsdate, oobservatory, minsec, tmplist, tmplist2
string obstime, timeobs, dateobs, century, val_imgets
real intime, ttime, mean, meanbad, x
int ncoadd, ddy, hhr, hhour, mmonth, yyear, dday, last2digits
bool ddarksub, pproject

iimagelist = imagelist
	list = iimagelist
	editlist = iimagelist
ddarksub = darksub
	if (ddarksub) {
		ddark = dark
		ddarkstring = darkstring
		darklist = iimagelist//ddarkstring
	}
bbadmask = badmask
pproject = project
	if (pproject) {
		tmplist = iimagelist//"_tMp"
		tmplist2 = iimagelist//"_tMp2"
		del (tmplist, ver-, >& "dev$null")
		del (tmplist2, ver-, >& "dev$null")
	}

oobservatory = observatory

# Add BPM and other goodies to the image headers, just like in the hadd.cl task
print("===> Adding various parameters to the image headers...")
list = iimagelist
while (fscan (list, current) !=EOF) {
  hedit(current,"BPM",bbadmask,add+,ver-,show+,update+)
  hedit(current,"BPMASK",bbadmask,add+,ver-,show+,update+)
  hedit(current,"ut",("(@'utime')"),add+,ver-,show+,update+)
  hedit(current,"st",("(@'sidtime')"),add+,ver-,show+,update+)
# Set exptime=itime * (# of coadds)
  imgets(current,"rdtmode", >& "dev$null")
  val_imgets=str(imgets.value)
  if(substr(val_imgets,1,5)!="COADD"){
    ncoadd=1.
  } else {
    ncoadd=real(substr(val_imgets,6,strlen(val_imgets)))
  }
  hedit(current,"coadds",(ncoadd),add+,ver-,show+,update+)
  imgets(current,"inttime")
  intime=real(imgets.value)
  hedit(current,"exptime",(ncoadd*intime),add+,ver-,show+,update+)
#
# The 'date-obs' keyword comes in either the old format "dd/mm/yy",
# the erroneous format "mm/dd/yy", or the new format "yyyy-mm-dd".
# Create new 'date-obs' keyword in the new format "yyyy-mm-dd".
  imgets(current,"dateorig",>&"dev$null")
  if (imgets.value!="0") {
    print ("DATEORIG keyword already exists -- cfhtfirst already run on image?")
  } else {
    imgets(current,"date-obs")
# Copy 'date-obs' keyword to 'dateorig' in original format.
    hedit(current,"dateorig",imgets.value,add+,ver-,show+,update+)
# Create header keyword 'obs-date', always in "dd/mm/yy" format.
    if(!ddmmyy){
# mm/dd/yy format:
      if(substr(imgets.value,3,3)=="/") {
        obsdate=substr(imgets.value,4,6)//substr(imgets.value,1,3)//substr(imgets.value,7,8)
      } else {
# yyyy-mm-dd format:
#  1234567890
        obsdate=substr(imgets.value,9,10)//"/"//substr(imgets.value,6,7)//"/"//substr(imgets.value,3,4)
      }
# create obs-date if date-obs in mm/dd/yy or yyyy-mm-dd format:
      hedit(current,"obs-date",obsdate,add+,ver-,show+,update+)
    } else {
# create obs-date if date-obs in dd/mm/yy format:
      hedit(current,"obs-date",imgets.value,add+,ver-,show+,update+)
    }
# Create new 'date-obs' in new format "yyyy-mm-dd" from 'obs-date' "dd/mm/yy"
#								    12345678
    imgets(current,"obs-date")
    if ((substr(imgets.value,7,7))=="9") {
  	  century=19
    } else {
  	  century=20
    }
    dateobs=century//substr(imgets.value,7,8)//"-"//substr(imgets.value,4,5)//"-"//substr(imgets.value,1,2)
    hedit(current,"date-obs",dateobs,add+,ver-,show+,update+)
  }
}

# Project from 3-D to 2-D image
if (pproject) {
	list = iimagelist 
	print("===> Projecting 3-D images down to 2-D...")
	while (fscan (list, name) !=EOF) {
	  fileroot(name,validim+)
	  base=fileroot.root
	  ext=fileroot.extension
	  if(ext!="")ext="."//ext
	  print(base//ext//"[*,*,1]", >> tmplist)
	  print(base//"_tMp2"//ext, >> tmplist2)
	}
	imcopy("@"//tmplist,"@"//tmplist2,ver-)
	imdel("@"//iimagelist,ver-)
	imren("@"//tmplist2,"@"//iimagelist,ver-)
	del(tmplist,ver-)
	del(tmplist2,ver-)
}

# Subtract off Dark Current 
if (ddarksub) {
	list = iimagelist 
# if darklist doesn't exist, assume dsub imgs don't either; create it & dsubs
	if(!access(darklist)) {
	    print("===> Subtracting off Dark Current...")
	    while (fscan (list, name) !=EOF) {
		fileroot(name,validim+)
		base=fileroot.root
		ext=fileroot.extension
		if(ext!="")ext="."//ext
		print(base//ddarkstring//ext, >> darklist)
  		imgets(name,"DARKSUB")
		if (imgets.value=="0") {
		    imarith(name, "-", ddark, base//ddarkstring//ext, verbose-)
	   hedit(base//ddarkstring//ext,"DARKSUB",ddark,add+,ver-,show+,update+)
		} else {
  print("DARKSUB keyword exists; cfhtfirst already run? "//name//" just copied")
		    imcopy(name, base//ddarkstring//ext, verbose-)
		}
	    }
	} else {
	    print("Darksubbed image list "//darklist//" already exists; images assumed to exist too.")
	}
#imarith("@"//iimagelist, "-", ddark, "@"//darklist, verbose-)
#hedit("@"//darklist,"DARKSUB",ddark,add+,ver-,show+,update+)
print(" ") 
print("===> Dark-subtracted images are listed in the file "//darklist//".") 
}

if(echog)beep

end

