# qksky.cl -- quick IR sky-subtraction routine using all images in a list
#
# This task creates a sky image from a list of flattened images
# and subtracts it from them. 
#
# This task should be run in the directory which contains the flattened images 
# listed in imagelist.  This script does not change those input images.  
# Previously a bad pixel image (mask) should have been produced.  
# The products of this routine are:
# --dark corrected, sky-subtracted, bad pixel zeroed frames: *.###substring.imh
# --a file (list.substring//rootname) which contains the names of those images
# --a sky image called rootname.sextension.imh
# --and the logfile.
#
# CHANGES:
# 940907: added parameter set for imcombine used to make the sky, as well as
#   option to use standard parameters for both object and std.star images
# 990217: added option to use iterstat; this slows things down if used but
#       produces a much better subtraction (note stdpars+ means iterstat-)
# 990218: added mksky & dosky options
# 990307: added addsky options
# 990320: added subpattern option
# 990406: fixed bugs of attempted double subtraction, and incorrect scaling,
#   plus changed ITERMEAN to ITERMED in scaling of sky image before subtraction
# 990506: added ZEROOUT keyword creation
# 991216: added check for ITERMED keyword and creation if absent
# 000702: imcombine "plfile" parameter now "nrejmasks", for V2.12 compatibility

procedure qksky(rootname,imagelist,logfile,substring,sextension,stdpars,stdstar,zeroout,maskname,tpl,tplname)

string rootname   {prompt="Rootname for sky image"}
string imagelist  {prompt="List of images to process (do not use an `@')"}
string logfile    {prompt="Name of log file"}
string substring  {"sq",prompt="String to be appended to sky-subtracted images"}
string sextension {"sq",prompt="Extension to be added to the sky image"}
bool   stdpars    {prompt="Use standard imcombine parameters to create sky?"}
bool   stdstar    {prompt="Are these standard star images?"}
bool   zeroout    {prompt="Zero out bad pixels on both sky & images?"}
string maskname   {prompt="Name of mask image to use"}
bool   tpl     {prompt="Divide by throughput correction after sky-subtraction?"}
string tplname {prompt="Name of throughput correction image to use"}
bool subpattern=no {prompt="Subtract pattern noise (before tpl)?"}
bool   mksky      {no,prompt="Make sky image even if it already exists?"}
bool   iterstat   {yes,prompt="Use iterative means when constructing sky?"}
bool   dosky	  {yes,prompt="Subtract sky from images once it's created?"}
bool   addsky	  {no,prompt="Add constant median sky level after subtraction?"}
bool   forcescale {no,prompt="Force recalculation of image medians?"}
bool   premask    {no,prompt="Mask when scaling before creating sky?"}
pset   imskypars  {prompt="Imcombine parameters"}
bool   echog      {yes,prompt="Beep when task is done?"}
struct *list
struct *datelist

begin

string rrootname, ddark, llogfile, bbadmask, eextension, ssextension
string skylist, sky, skybase, meanfile, currentsky, base, mmaskname
string iimagelist, SkyImageList, ssubstr, check, ttplname, subpat_list
string datefile, thedate, ickinfo, name, img, img2, joinlist
string s_sigma, s_plfile, s_combine, s_reject, s_offsets, s_masktyp
string s_scale, s_weight, s_statsec, s_expname, s_zero, s_rdnoise, s_gain
real s_maskval, s_blank, s_lthresh, s_hthresh, s_nlow, s_nhigh, s_nkeep
real s_lsigma, s_hsigma, s_pclip, s_grow, s_sigscale
real mean, meanbad, x, valmed, valsky
int sstartno, eendno, i, j, numpix
bool zzeroout, ttpl, sstdpars, sstdstar, iter_stat, scalecalc, force_scale, go
bool pre_mask, mk_sky, do_sky, add_sky, sub_pattern, headadd

rrootname = rootname
iimagelist = imagelist
	list = iimagelist
llogfile = logfile
ssubstr = substring
ssextension = sextension
sstdpars = stdpars
sstdstar = stdstar
zzeroout = zeroout
mmaskname = maskname
ttpl = tpl
ttplname = tplname

sub_pattern = subpattern
mk_sky = mksky
iter_stat = iterstat
do_sky = dosky
add_sky = addsky
force_scale = forcescale
pre_mask = premask

s_sigma = i_sigma
s_plfile = i_plfile
s_combine = i_combine
s_reject = i_reject
s_masktyp = i_masktyp
s_maskval = i_maskval
s_blank = i_blank
s_scale = i_scale
s_zero = i_zero
s_weight = i_weight
s_statsec = i_statsec
s_expname = i_expname
s_lthresh = i_lthresh
s_hthresh = i_hthresh
s_nlow = i_nlow
s_nhigh = i_nhigh
s_nkeep = i_nkeep
s_lsigma = i_lsigma
s_hsigma = i_hsigma
s_rdnoise = i_rdnoise
s_gain = i_gain
s_sigscale = i_sigscale
s_pclip = i_pclip
s_grow = i_grow

# The "go" parameter won't let the task run if the needed tasks aren't loaded.
go=yes
if(iter_stat && !defpac("imgtools")) {go=no}
if(!iter_stat && add_sky) {go=no}
if(!iter_stat && sub_pattern) {go=no}
if(go) {

# Set up log file, appending if it already exists
print("Let the games begin...")
if(access(llogfile))
  print("Log file `"//llogfile//"' already exists... will append to it.")
print(" ",>>llogfile);print("PHIIRS.QKSKY--------------------------",>>llogfile)

# Set up files
sky = rrootname//"."//ssextension   # the final sky image
SkyImageList = iimagelist//ssubstr  # @file for sky-subtracted images
if(access(SkyImageList)) delete(SkyImageList,ver-)
while (fscan (list, name) !=EOF) {
  fileroot(name,validim+)
  check=fileroot.extension
  if(check!="")check="."//check
  base=fileroot.root
  print(base//ssubstr//check, >> SkyImageList)
}

# Make sky, if desired
if (mk_sky || !access(sky)) {
print("===> Making sky...")
if (access(sky)) imdel (sky, ver-)
print("File "//SkyImageList//" contains the following filenames:", >> llogfile)
type(SkyImageList, >> llogfile)
print(" ", >> llogfile)

# Calculate scalings for all frames, if necessary
if (iter_stat) {
    print("===> Calculating scalings using iterstat (or maskstat)...")
    list = iimagelist
# Calculate the median of the image.  If premask=yes, use only unmasked pixels.
# Record the median value into the image header with the card SKYMED.
# If SKYMED card already exists in the header, do not recalculate
# unless the parameter forcescale=yes.
    while (fscan (list, img) !=EOF) {
	scalecalc = force_scale
	headadd = no
	imgets (img,"ITERMED",>&"dev$null")
	if (imgets.value == "0") headadd = yes
	if (!force_scale) {
		imgets (img,"SKYMED",>&"dev$null")
		if (imgets.value == "0") scalecalc = yes
		imgets (img,"ITERMED",>&"dev$null")
		if (imgets.value == "0") scalecalc = yes
	}
	if (scalecalc) {
	    if(pre_mask) {
		maskstat(img,maskname,1.,nsigrej=5.,maxiter=10,lower=INDEF,upper=INDEF,iterstat+,addhead=headadd,ver-,print-)
	    } else {
         	iterstat(img,nsigrej=5.,maxiter=10,lower=INDEF,upper=INDEF,print-,ver-,addhead=headadd)
	    }
	    valmed = 1./iterstat.median
	    hedit (img,"SKYMED",valmed,add+,ver-,show-,update+)
	}
    }
}

# Combine all images to make the sky
print(" ", >> llogfile)
print("===> Combine all images to make the sky")
print("===> Combine all images to make the sky", >> llogfile)
if(sstdpars){
	if(sstdstar){
		imcombine("@"//iimagelist, sky, sigma=s_sigma, nrejmasks=s_plfile, combine="median", reject="ccdclip", project-, offsets="none", masktype="none", maskvalue=s_maskval, blank=s_blank, scale="none", zero="none", weight="none", statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=1, mclip+, lsigma=3, hsigma=3, rdnoise=s_rdnoise, gain=s_gain, sigscale=s_sigscale, pclip=s_pclip, grow=s_grow, logfile='STDOUT', >> llogfile)
	} else {
		imcombine("@"//iimagelist, sky, sigma=s_sigma, nrejmasks=s_plfile, combine="median", reject="ccdclip", project-, offsets="none", masktype="none", maskvalue=s_maskval, blank=s_blank, scale="median", zero="none", weight="none", statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=1, mclip+, lsigma=3, hsigma=3, rdnoise=s_rdnoise, gain=s_gain, sigscale=s_sigscale, pclip=s_pclip, grow=s_grow, logfile='STDOUT', >> llogfile)
	}
} else {
	if (iter_stat) {
		imcombine("@"//iimagelist, sky, sigma=s_sigma, nrejmasks=s_plfile, combine=s_combine, reject=s_reject, project-, offsets="none", masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale="!SKYMED", zero=s_zero, weight=s_weight, statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, sigscale=s_sigscale, pclip=s_pclip, grow=s_grow, logfile='STDOUT', >> llogfile)
	} else {
		imcombine("@"//iimagelist, sky, sigma=s_sigma, nrejmasks=s_plfile, combine=s_combine, reject=s_reject, project-, offsets="none", masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale=s_scale, zero=s_zero, weight=s_weight, statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, sigscale=s_sigscale, pclip=s_pclip, grow=s_grow, logfile='STDOUT', >> llogfile)
	}
}
# Make sure ITERMED header card isn't carried over from the first input image
hedit(sky, "ITERMED", del+, add-, ver-, show-, update+)

if(zzeroout){
	imar(sky, "*", mmaskname, sky, verbose-)
	hedit(sky,"ZEROOUT",mmaskname,del-,add+,ver-,sho-,upd+)
	print("===> Zeroed out bad pixels on the sky image", >> llogfile)
} else {
	print("===> Bad pixel zeroing not done on the sky image", >> llogfile)
}
print(" ", >> llogfile)

# End of mksky+ portion
} else {
 print("===> "//sky//" already exists -- not recreated -- continuing...")
 print(sky//" already exists -- not recreated",>>llogfile);print(" ",>>llogfile)
}

if (iter_stat) {
	imgets (sky,"ITERMED",>&"dev$null")
	valsky = real(imgets.value)
	if(pre_mask) {
		if (imgets.value=="0" || force_scale) {
			maskstat(sky,maskname,1.,nsigrej=5.,maxiter=10,lower=INDEF,upper=INDEF,iterstat+,ver-,print+, >> llogfile)
			valsky = iterstat.median
		}
	} else {
		if (imgets.value=="0" || force_scale) {
			iterstat(sky,nsigrej=5.,maxiter=10,lower=INDEF,upper=INDEF,print+,ver-, >> llogfile)
			valsky = iterstat.median
		}
	}
} else {
	imstat(sky, >> llogfile)
}

if (do_sky) {
# Subtract the Sky from all Images
print("===> Subtract the Sky from all Images ")
print("",>> llogfile);print("===> Subtract the Sky from all Images", >>llogfile)
if (add_sky) {
	print("===> Then Add Back In Median Sky Level (addsky+)")
	print("===> Then Add Back In Median Sky Level (addsky+)", >>llogfile)
}
if (iter_stat) {
	joinlist=mktemp("join")
	join (iimagelist, SkyImageList, >> joinlist)
	list = joinlist
	while (fscan (list, img, img2) !=EOF) {
# Scale sky image
		imgets (img,"ITERMED",>&"dev$null")
		valmed=real(imgets.value)/valsky
# Subtract sky image & pattern noise, & add back in sky level if desired.
		if (sub_pattern) {
		    print("===> Subtracting pattern noise from all images ")
	   print("===> Subtracting pattern noise from all images ", >> llogfile)
		    subpat_list = mktemp("subpat")
                    print (img2, >> subpat_list)
                    if (add_sky) {
                        imdel ("addsky_pattern.fits",ver-,>& "dev$null")
                        if (!access("blank_pattern.fits")) imar (subimg, "-", subimg, "blank_pattern.fits", pix="real", calc="real", ver-)
                        imar ("blank_pattern.fits", "-", valmed, "addsky_pattern.fits", pix="real", calc="real", ver-)
                        subpattern (subpat_list, "", startline=1, finishline=512, blankname="addsky_pattern.fits", ver-, useimgs+, saveimgs-)
                        imdel ("addsky_pattern.fits",ver-,>& "dev$null")
                    } else {
                        subpattern (subpat_list, "", startline=1, finishline=512, blankname="blank_pattern.fits", ver-, useimgs+, saveimgs-)
                    }
                    del (subpat_list, ver-)
		    print(" ", >> llogfile)
		} else {
# Don't subtract pattern noise, but subtract sky & poss. add sky level back in.
		    if (add_sky) {
		 imcalc(img//","//sky,img2,"im1-im2*"//valmed//"+"//valmed,ver-)
		    } else {
			imcalc(img//","//sky,img2,"im1-im2*"//valmed,ver-)
		    }
		}
	}
	del(joinlist,ver-,>&"dev$null")
} else {
	imarith("@"//iimagelist, "-", sky, "@"//SkyImageList, verbose-)
      print("IMARITH: @"//iimagelist//" - "//sky//" @"//SkyImageList,>>llogfile)
	print(" ", >> llogfile)
}

# Make throughput correction, if desired
if(ttpl){
	print("===> Make throughput correction on all images ")
	print("===> Make throughput correction on all images ", >> llogfile)
	imarith("@"//SkyImageList, "/", ttplname, "@"//SkyImageList, verbose-)
	print("IMARITH: @"//SkyImageList//" / "//ttplname//" @"//SkyImageList, >> llogfile)
	print(" ", >> llogfile)
}

# Force updating of pixel values, just to be safe
print("===> Updating pixel values... ")
minmax("@"//SkyImageList, force+, update+, verbose-)
if (!iter_stat) imstat("@"//SkyImageList, >> llogfile)

# End of do_sky=yes option
} else {
print("===> Sky made but NOT subtracted from images... ")
}

print("===> Wrapping up... ")
datefile = mktemp("date")
date(>>datefile)
datelist = datefile
print(" ", >> llogfile); print("finished at "//datelist, >> llogfile)
datelist = ""
del(datefile, ver-)
print("...it is accomplished.")

} else {
	if(!iter_stat && add_sky) {
	    print("WARNING: iterstat=no & addsky=yes are incompatible options!")
	} else if (!iter_stat && sub_pattern) {
       print("WARNING: iterstat=no & sub_pattern=yes are incompatible options!")
	} else {
	    print("WARNING: stsdas.toolbox.imgtools not loaded (for imcalc)!")
	}
}

if(echog)beep

end

